/* url_gj_mpi.c
 * Rozwiazywanie ukladow rownan liniowych metoda eliminacji
 * Gaussa-Jordana
 * @author Piotr Zarowny, Politechnika Warszawska
 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/time.h>
#include <mpi.h>

double** A = NULL;  // macierz A
double*  b = NULL;  // wektor b
double*  x = NULL;  // wektor x
int      n = 0;     // wymiar macierzy/wektorow

int procId, procCnt;             // procesor, liczba procesorow
int ni, procSrc;                 // rozmiar przedzialu na pojedynczym
                                 // procesorze
MPI_Comm comm = MPI_COMM_WORLD;  // komunikator

FILE *fx;           // plik wynikow
clock_t ct;         // czas dzialania programu

int i, j, k;        // zmienne pomocnicze
int ib, ik;
double alfa;

int main(int argc, char* argv[]) {

    MPI_Init(&argc, &argv);
    MPI_Comm_rank(comm, &procId);
    MPI_Comm_size(comm, &procCnt);
                                              // pobranie wymiaru n
    if (procId == 0) {
        if (argc == 2) {
            sscanf(argv[1], "%d", &n);
        } else {
            printf("\nProsze podac wymiar n: ");
            scanf("%d", &n);
        }
        ct = clock();
    }
    MPI_Bcast(&n, 1, MPI_INT, 0, comm);

    if (n <= 0) {
        if (procId == 0)
            printf("\nBLAD: podano bledna wartosc wymiaru n!\n");
        MPI_Finalize();
        exit(0);
    }
    if (procId == 0)
        printf("\nObliczanie... ");
                                              // utworzenie A, b, x
    if ((A = (double **)malloc((n + 1) * sizeof(double *))) == NULL) {
        printf("\nBLAD: nie utworzono macierzy A!\n");
        MPI_Finalize();
        exit(1);
    }
    for (i = 0; i <= n; i++)
       A[i] = (double *)malloc((n + 1) * sizeof(double));

    if ((b = (double *)malloc((n + 1) * sizeof(double))) == NULL) {
        printf("\nBLAD: nie utworzono wektora b!\n");
        MPI_Finalize();
        exit(1);
    }
    if ((x = (double *)malloc((n + 1) * sizeof(double))) == NULL) {
        printf("\nBLAD: nie utworzono wektora x!\n");
        MPI_Finalize();
        exit(1);
    }
                                              // wypelnienie A, b, x
    for (i = 1; i <= n; i++) {
        for (j = 1; j <= n; j++)
            A[j][i] = (i != j) ? pow(-1,i + j - 1) / (i + j - 1) :
	                                     2 + pow(-1, i) / pow(i, 2);

       b[i] = 3 + pow(-1, i + 1) * sin(i);
       x[i] = 0;
    }
                                              // podzial i na przedzialy
    ni = (int)(ceil((float)n / (float)procCnt));
    ib = procId * ni + 1;
    ik = ib + ni - 1;
    if (procId == procCnt - 1)
        ik = n;
                                  // algorytm eliminacji Gaussa-Jordana
    for (k = 1; k <= n; k++) {
        for (i = ib; i <= ik; i++) {
            if (i == k)
                continue;

            alfa = A[i][k] / A[k][k];

            for (j = k; j <= n; j++)
                A[i][j] = A[i][j] - alfa * A[k][j];

            b[i] = b[i] - alfa * b[k];
        }
                                // rozeslanie wiersza eliminujacego
       if (k < n) {
           procSrc = k/ni;
           MPI_Bcast(A[k + 1], n + 1, MPI_DOUBLE, procSrc, comm);
           MPI_Bcast(&b[k + 1], 1, MPI_DOUBLE, procSrc, comm);
       }
    }
    for (k = ib; k <= ik; k++)
        x[k] = b[k] / A[k][k];
                                          // zebranie wszystkich wynikow
    MPI_Gather(&x[ib], ni, MPI_DOUBLE, &x[ib], ni, MPI_DOUBLE, 0, comm);

                                              // zapis wynikow do pliku
    if (procId == 0) {
        if ((fx = fopen("wyniki.txt", "w")) != NULL) {
            for (i = 1; i <= n; i++)
                fprintf(fx,"%.4f\n",x[i]);

            fclose(fx);
        } else
            printf("\nBLAD: nie zapisano wynikow do pliku!\n");
    }
                                              // czyszczenie pamieci
    for(i = 0; i <= n; i++)
        free(A[i]);
    free(A);
    free(b);
    free(x);
                                              // zakonczenie programu
    if (procId == 0) {
        printf("OK\nCzas wykonania: %.2f [s]\n",
                                ((double)(clock() - ct) / CLOCKS_PER_SEC));
    }
    MPI_Finalize();
    return 0;
}

